#include <OpenSim/OpenSim.h>
#include "OpenSimPendulum.h"
#include <sstream>
#include <string>
#include <cstdio>
#include <vector>






OpenSimPendulum::OpenSimPendulum( int numberOfLinks, bool verboseMessages)
{


    try{

        model = new OpenSim::Model;

        if(verboseMessages) 
            printf("createOpenSimPendulum:\n");

        numberOfPendulumLinks = numberOfLinks;
        
        double gravityValue     = 9.81;
        double massValue        = 1;
        double inertiaValue     = 1;
        double pendulumLength   = 1;

        if(verboseMessages) 
            printf("\t:Adding gravity\n");

        model->set_gravity(SimTK::Vec3(0,-gravityValue,0));
        std::stringstream objName;

        bodyVector.resize(numberOfPendulumLinks);
        jointVector.resize(numberOfPendulumLinks);
        muscleVector.resize(numberOfPendulumLinks);

        //const OpenSim::Ground& groundFrame = model.getGround();

        for (int i = 0; i < numberOfPendulumLinks; ++i) {

            objName.str("");
            objName << "link_" << i;

            bodyVector.at(i) = new OpenSim::Body(objName.str(),
                                    massValue,
                                    SimTK::Vec3(0),
                                    SimTK::Inertia(inertiaValue));

            objName.str("");
            objName << "muscle_" << i;

            muscleVector.at(i) = new OpenSim::Millard2012EquilibriumMuscle(
                            objName.str(),
                            20,
                            pendulumLength*0.9,
                            pendulumLength*0.1,
                            0);


            objName.str("");
            objName << "joint_" << i;

            if(i==0){
                jointVector.at(i) =  new OpenSim::PinJoint(objName.str(),
                                               model->getGround(),
                                               SimTK::Vec3(0),
                                               SimTK::Vec3(0),
                                               *bodyVector.at(0),
                                               SimTK::Vec3(0, pendulumLength, 0),
                                               SimTK::Vec3(0));
                objName.str("p");
                objName << i;
                muscleVector.at(i)->addNewPathPoint(objName.str(),
                                        *bodyVector.at(0),
                                        SimTK::Vec3(0.01,
                                                    pendulumLength/2.0,
                                                    0.));

                objName.str("pG");
                muscleVector.at(i)->addNewPathPoint(objName.str(),
                                        model->updGround(),
                                        SimTK::Vec3(0.01,
                                                    pendulumLength/2.0,
                                                    0.));



            }else{
                jointVector.at(i) = new OpenSim::PinJoint(objName.str(),
                                               *bodyVector.at(i-1),
                                               SimTK::Vec3(0),
                                               SimTK::Vec3(0),
                                               *bodyVector.at(i),
                                               SimTK::Vec3(0, pendulumLength, 0),
                                               SimTK::Vec3(0));

                objName.str("p");
                objName << i;
                muscleVector.at(i)->addNewPathPoint(objName.str(),
                                        *bodyVector.at(i),
                                        SimTK::Vec3(0.01,
                                                    pendulumLength/2.0,
                                                    0.));
                objName.str("p");
                objName << i-1;
                muscleVector.at(i)->addNewPathPoint(objName.str(),
                                        *bodyVector.at(i-1),
                                        SimTK::Vec3(0.01,
                                                    pendulumLength/2.0,
                                                    0.));
            }


            model->addBody(bodyVector.at(i));
            model->addJoint(jointVector.at(i));
            model->addForce(muscleVector.at(i));

            if(verboseMessages) 
                printf("\t:Adding link %i\n",i);
        }

        if(verboseMessages) printf("\t:getDefaultState\n");

        *modelState = model->initSystem();

        for(int i=0; i<numberOfPendulumLinks; ++i){
            model->updCoordinateSet()[i].setValue(*modelState,
                                             0*SimTK::Pi);
            OpenSim::Millard2012EquilibriumMuscle& muscle =
           (OpenSim::Millard2012EquilibriumMuscle&)model->updForceSet()
                                                         .updMuscles()[i];
            muscle.setIgnoreTendonCompliance(*modelState, true);
            muscle.setIgnoreActivationDynamics(*modelState, true);
        }

        model->equilibrateMuscles(*modelState);


        if(verboseMessages){
            printf("\t:Number of model states"
                    ": NU: %i, NQ: %i, NZ: %i\n",
                     modelState->getNU(),
                     modelState->getNQ(),
                     modelState->getNZ());
        /*Note:
            U: generalized velocities
            Q: generalized positions
            Z: auxiliary states (non-multibody states)
        */

        }
        




    }catch(const std::exception& e) {
        std::cout << "EXCEPTION: createOpenSimPendulum: " 
                  << e.what() << std::endl;

    }


}



void OpenSimPendulum::calcOpenSimPendulumStateDerivative(
             double time,
             const std::vector< double >& generalizedPositions,
             const std::vector< double >& generalizedVelocities,
             const std::vector< double >& muscleActivations,
             std::vector< double >& generalizedAcceleration,
             bool verboseMessages)
{


    if(verboseMessages)
        printf("createStateDerivative:\n");


    if(verboseMessages){
        printf("position\tvelocity\tactivation\n");
        for(int i=0; i<numberOfPendulumLinks;++i){
            printf("%e\t%e\t%e\n",
                   generalizedPositions.at(i),
                   generalizedVelocities.at(i),
                   muscleActivations.at(i));
        }
    }

    if(verboseMessages)
        printf("\tCopying state over\n");




    for(int i=0; i<numberOfPendulumLinks;++i){




        modelState->updU()[i] = generalizedVelocities.at(i);
        modelState->updQ()[i] = generalizedPositions.at(i);

        OpenSim::Millard2012EquilibriumMuscle& muscle =
       (OpenSim::Millard2012EquilibriumMuscle&)model->updForceSet()
                                                    .updMuscles()[i];

        muscle.setActivation(*modelState,muscleActivations.at(i));

    }


    if(verboseMessages)
        printf("\trealize to acceleration\n");
    model->realizeAcceleration(*modelState);

    if(verboseMessages)
        printf("\tgetUDot\n");

    generalizedAcceleration.resize(numberOfPendulumLinks);

    for(int i=0; i<numberOfPendulumLinks;++i){
        generalizedAcceleration.at(i) = modelState->getUDot()[i];
    }


    if(verboseMessages)
        for(int i=0; i<modelState->getUDot().nelt(); i++)
              printf("\t%i: %f\n",i, modelState->getUDot()[i] );


}



